from odoo.tests import tagged, Form
from odoo.addons.service_purchase.tests.common import TestCommonSalePurchaseNoChart


@tagged('-at_install', 'post_install')
class TestSalePurchase(TestCommonSalePurchaseNoChart):

    @classmethod
    def setUpClass(cls, chart_template_ref=None):
        super().setUpClass(chart_template_ref=chart_template_ref)
        cls.partner_a.country_id = cls.env.ref('base.be')

        # create a generic Sale Order with 2 classical products and a purchase service
        SaleOrder = cls.env['sale.order'].with_context(tracking_disable=True)
        cls.sale_order_1 = SaleOrder.create({
            'partner_id': cls.partner_a.id,
            'partner_invoice_id': cls.partner_a.id,
            'partner_shipping_id': cls.partner_a.id,
            'pricelist_id': cls.company_data['default_pricelist'].id,
        })

        cls.sol1_service_purchase_1 = cls.env['sale.order.line'].create({
            'product_id': cls.service_purchase_2.id,
            'product_uom_qty': 4,
            'order_id': cls.sale_order_1.id,
            'tax_id': False,
        })

        cls.sale_order_2 = SaleOrder.create({
            'partner_id': cls.partner_a.id,
            'partner_invoice_id': cls.partner_a.id,
            'partner_shipping_id': cls.partner_a.id,
            'pricelist_id': cls.company_data['default_pricelist'].id,
            'company_id': cls.company_data['company'].id,
        })
        cls.sol2_service_purchase_2 = cls.env['sale.order.line'].create({
            'product_id': cls.service_purchase_1.id,
            'product_uom_qty': 7,
            'order_id': cls.sale_order_2.id,
            'tax_id': False,
        })

        cls.sale_order_3 = SaleOrder.create({
            'partner_id': cls.partner_a.id,
            'partner_invoice_id': cls.partner_a.id,
            'partner_shipping_id': cls.partner_a.id,
            'pricelist_id': cls.company_data['default_pricelist'].id,
            'company_id': cls.company_data_2['company'].id,
        })
        cls.sol3_service_purchase_1 = cls.env['sale.order.line'].create({
            'product_id': cls.service_purchase_1.id,
            'product_uom_qty': 7,
            'order_id': cls.sale_order_3.id,
            'tax_id': False,
        })
        cls.sol3_service_purchase_3 = cls.env['sale.order.line'].create({
            'product_id': cls.service_purchase_3.id,
            'product_uom_qty': 7,
            'order_id': cls.sale_order_3.id,
            'tax_id': False,
        })

    def confirm_and_pay(self, so):
        so.action_confirm()
        inv = so._create_invoices()
        inv.action_post()
        action_data = inv.action_register_payment()
        wizard = Form(self.env['account.payment.register'].with_context(action_data['context'])).save()
        action = wizard.action_create_payments()
        return self.env['account.payment'].browse(action['res_id']).move_id

    def unreconcile_payment(self, invoice):
        info, = invoice._get_all_reconciled_invoice_partials()
        apr = self.env['account.partial.reconcile'].browse(info['partial_id'])
        if apr.debit_move_id.id not in invoice.line_ids.ids:
            payment_id = apr.debit_move_id.id
        else:
            payment_id = apr.credit_move_id.id

        invoice.js_remove_outstanding_partial(apr.id)

        return payment_id

    def reconcile_payment(self, invoice, payment_id):
        invoice.js_assign_outstanding_line(payment_id)

    def test_sale_create_purchase(self):
        po_sudo = self.env['purchase.order'].sudo()
        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 0, "Start without PO open")

        invoice1 = self.confirm_and_pay(self.sale_order_1)

        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 1, "Only one PO should have been created for this partner.")
        self.assertEqual(len(purchase_order_BC.order_line), 1, "The purchase order should have 1 line.")

        payment1 = self.unreconcile_payment(invoice1)

        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 1, "Only one PO should have been created for this partner.")
        self.assertEqual(len(purchase_order_BC.order_line), 1, "Unreconciliation of payment should not remove a PO line.")

        self.reconcile_payment(invoice1, payment1)

        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 1, "Only one PO should have been created for this partner.")
        self.assertEqual(len(purchase_order_BC.order_line), 1, "Re-reconciliation of payment should not create a new PO line.")

        self.confirm_and_pay(self.sale_order_3)

        purchase_order_A = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceA.id), ('state', '=', 'draft')])
        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_A), 1, "Only one PO should have been created for this partner.")
        self.assertEqual(len(purchase_order_BC), 1, "Only one PO should have been created for this partner.")
        self.assertEqual(len(purchase_order_BC.order_line), 2, "The purchase order should have 2 lines.")
        self.assertEqual(len(purchase_order_A.order_line), 1, "The purchase order should have 1 line.")

        self.confirm_and_pay(self.sale_order_2)

        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 1, "Only one PO should have been created for this partner.")
        self.assertEqual(len(purchase_order_BC.order_line), 2, "The purchase order should have 2 lines")

        self.assertEqual(
            purchase_order_A.company_id,
            self.env.ref('base.main_company'),
            "Should be on Main company whatever the company of invoice / sale."
        )
        self.assertEqual(
            purchase_order_BC.company_id,
            self.env.ref('base.main_company'),
            "Should be on Main company whatever the company of invoice / sale."
        )

    def test_sale_reuse_purchase(self):
        """
        Ensure that, even if a partner has a distinct commercial_partner_id,
        the PO lines will be created on an existing draft PO associated to this partner
        """
        po_sudo = self.env['purchase.order'].sudo()

        # Set the partner's commercial_partner_id to itself
        self.partner_vendor_serviceBC.is_company = True
        self.confirm_and_pay(self.sale_order_1)

        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBC.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 1, "Exactly one PO should have been created for the partner.")
        self.assertEqual(len(purchase_order_BC.order_line), 1, "The purchase order should have 1 line.")

        # Reset the partner's commercial_partner_id to another partner
        self.partner_vendor_serviceBC.is_company = False
        self.confirm_and_pay(self.sale_order_3)

        # The PO created for the partner should be reused
        purchase_order_BCcpy = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBCcpy.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BCcpy), 0, "No PO should have been created for the company.")

        purchase_order_BC = po_sudo.search([('partner_id', '=', self.partner_vendor_serviceBC.id), ('state', '=', 'draft')])
        self.assertEqual(len(purchase_order_BC), 1, "Exactly one PO should have been created for the company.")
        self.assertEqual(len(purchase_order_BC.order_line), 2, "The purchase order should have 2 line.")
